import unittest
import numpy as np
from lightwood.analysis.nc.norm import Normalizer


class TestNcNormalizer(unittest.TestCase):
    def test_compute_numerical_labels(self):
        preds = np.array([1, 2.1, 3.2, -2, 1e10, 0])
        truths = np.array([1, 2, 3, 4, 5, 0])
        bounds = [0.5, 1.5]

        labels = Normalizer.compute_numerical_labels(preds, truths, bounds)
        self.assertTrue(np.min(labels) >= bounds[0])
        self.assertTrue(np.max(labels) <= bounds[1])
        self.assertFalse(np.isnan(labels).any())
        self.assertFalse(np.isinf(labels).any())

    def test_compute_categorical_labels(self):
        preds = np.array([[0.01, 0.99], [1, 0], [0.99, 0.1], [0, 1]])
        truths = np.array([[0, 1], [0, 1], [1, 0], [1, 0]])

        labels = Normalizer.compute_categorical_labels(preds, truths)
        self.assertTrue(np.min(labels) >= 0)
        self.assertFalse(np.isnan(labels).any())
        self.assertFalse(np.isinf(labels).any())



import unittest
import pandas as pd
from lightwood.analysis import ShapleyValues
from lightwood.api.high_level import ProblemDefinition, json_ai_from_problem
from lightwood.api.high_level import code_from_json_ai, predictor_from_code


class TestSHAP(unittest.TestCase):
    def test_0_shap_analysis(self):
        if ShapleyValues is None:
            print('Skipping this test since the Shapley values library is not installed')
            return

        df = pd.read_csv('tests/data/hdi.csv')
        target = 'Development Index'

        pdef = ProblemDefinition.from_dict({'target': target, 'time_aim': 40})
        json_ai = json_ai_from_problem(df, problem_definition=pdef)

        json_ai.analysis_blocks = [{
            'module': 'lightwood.analysis.ShapleyValues',
            'args': {}
        }]

        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)

        predictor.learn(df)
        predictions = predictor.predict(df.head())

        self.assertIn('shap_explainer', predictor.runtime_analyzer)

        self.assertIn('shap_base_response', predictions.columns)
        self.assertIn('shap_final_response', predictions.columns)
        for input_col in df.columns:
            if input_col != target:
                self.assertIn(f'shap_contribution_{input_col}', predictions.columns)



from copy import deepcopy
from typing import Dict, Union

import numpy as np
import pandas as pd
import mxnet as mx

from gluonts.dataset.pandas import PandasDataset

from gluonts.model.deepar import DeepAREstimator  # @TODO: support for other estimators
from gluonts.mx import Trainer
from gluonts.mx.trainer.callback import TrainingHistory

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class GluonTSMixer(BaseMixer):
    horizon: int
    target: str
    supports_proba: bool
    model_path: str
    hyperparam_search: bool
    default_config: dict

    def __init__(
            self,
            stop_after: float,
            target: str,
            horizon: int,
            window: int,
            dtype_dict: Dict,
            ts_analysis: Dict,
            n_epochs: int = 10,
            early_stop_patience: int = 3,
            seed: int = 0,
    ):
        """
        Wrapper around GluonTS probabilistic deep learning models. For now, only DeepAR is supported.

        :param stop_after: time budget in seconds.
        :param target: column to forecast.
        :param horizon: length of forecasted horizon.
        :param window: length of input data.
        :param dtype_dict: data type of each column in the dataset.
        :param ts_analysis: dictionary with miscellaneous time series info, as generated by 'lightwood.data.timeseries_analyzer'.
        :param n_epochs: amount of epochs to train the model for. Will perform early stopping automatically if validation loss degrades.
        :param early_stop_patience: amount of consecutive epochs with no improvement in the validation loss.
        :param seed: specifies the seed used internally by GluonTS for reproducible predictions
        """  # noqa
        super().__init__(stop_after)
        self.stable = True
        self.prepared = False
        self.supports_proba = False
        self.target = target
        self.window = window
        self.horizon = horizon
        self.n_epochs = n_epochs
        self.dtype_dict = dtype_dict
        self.ts_analysis = ts_analysis
        self.grouped_by = ['__default'] if not ts_analysis['tss'].group_by else ts_analysis['tss'].group_by
        self.model = None
        self.train_cache = None
        self.patience = early_stop_patience
        self.seed = seed

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """ Fits the model. """  # noqa
        log.info('Started fitting GluonTS forecasting model')

        # prepare data
        cat_ds = ConcatedEncodedDs([train_data, dev_data])
        train_ds = self._make_initial_ds(cat_ds.data_frame, train=True)

        estimator = DeepAREstimator(
            freq=train_ds.freq,
            prediction_length=self.horizon,
            trainer=Trainer(epochs=self.n_epochs, callbacks=[EarlyStop(patience=self.patience)])
        )
        self.model = estimator.train(train_ds)
        log.info('Successfully trained GluonTS forecasting model.')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Due to how lightwood implements the `update` procedure, expected inputs for this method are:

        :param dev_data: original `test` split (used to validate and select model if ensemble is `BestOf`).
        :param train_data: concatenated original `train` and `dev` splits.
        """  # noqa
        self.hyperparam_search = False
        self.fit(dev_data, train_data)
        self.prepared = True

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """ 
        Calls the mixer to emit forecasts.
        """  # noqa
        length = sum(ds.encoded_ds_lenghts) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=['prediction', 'lower', 'upper'],
                           dtype=object)
        ydf['index'] = ds.data_frame.index
        conf = args.fixed_confidence if args.fixed_confidence else 0.9
        ydf['confidence'] = conf

        gby = self.ts_analysis["tss"].group_by if self.ts_analysis["tss"].group_by else []
        groups = ds.data_frame[gby[0]].unique().tolist() if gby else None

        for idx in range(length):
            df = ds.data_frame.iloc[:idx] if idx != 0 else None
            input_ds = self._make_initial_ds(df, groups=groups)
            if not input_ds:
                # edge case: new group
                for col in ['prediction', 'lower', 'upper']:
                    ydf.at[idx, col] = [0 for _ in range(self.ts_analysis["tss"].horizon)]
            else:
                mx.random.seed(self.seed)
                np.random.seed(self.seed)
                forecasts = list(self.model.predict(input_ds))[0]
                ydf.at[idx, 'prediction'] = [entry for entry in forecasts.quantile(0.5)]
                ydf.at[idx, 'lower'] = [entry for entry in forecasts.quantile(1 - conf)]
                ydf.at[idx, 'upper'] = [entry for entry in forecasts.quantile(conf)]

        return ydf

    def _make_initial_ds(self, df=None, train=False, groups=None):
        oby = self.ts_analysis["tss"].order_by
        gby = self.ts_analysis["tss"].group_by if self.ts_analysis["tss"].group_by else []
        freq = self.ts_analysis['sample_freqs']['__default']
        keep_cols = [f'__mdb_original_{oby}', self.target] + [col for col in gby]

        if df is None and not train:
            df = self.train_cache
            if gby:
                df = df[df[gby[0]].isin(groups)]
        else:
            sub_df = df[keep_cols]
            df = deepcopy(sub_df)

            if train:
                self.train_cache = df
            else:
                if gby:
                    cache = self.train_cache[self.train_cache[gby[0]].isin(groups)]
                else:
                    cache = self.train_cache
                df = pd.concat([cache, df]).sort_index()

        df = df.drop_duplicates()  # .reset_index(drop=True)

        if len(df) == 0:
            return None

        if gby:
            df = df.groupby(by=gby[0]).resample(freq).sum().reset_index(level=[0])
            # @TODO: multiple group support and remove groups without enough data
        else:
            df = df.resample(freq).sum()
            gby = '__default_group'
            df[gby] = '__default_group'

        ds = PandasDataset.from_long_dataframe(df, target=self.target, item_id=gby, freq=freq)
        return ds


class EarlyStop(TrainingHistory):
    def __init__(self, patience=3):
        super().__init__()
        self.patience = max(1, patience)
        self.counter = 0

    def on_validation_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network,
        trainer,
    ) -> bool:
        super().on_validation_epoch_end(epoch_no, epoch_loss, training_network, trainer)

        if len(self.validation_loss_history) > 1:
            if self.validation_loss_history[-1] > self.validation_loss_history[-2]:
                self.counter += 1
            else:
                self.counter = 0  # reset if not successive

        if self.counter >= self.patience:
            return False
        else:
            return True



import unittest
from lightwood.encoder.categorical import CategoricalAutoEncoder

import string
import random
import logging

import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score
import torch

from lightwood.helpers.log import log


class TestAutoencoder(unittest.TestCase):

    def create_test_data(self,
                         nb_categories=500,
                         nb_int_categories=50,
                         max_category_size=50,
                         test_data_rel_size=0.33):
        random.seed(2)
        np_random = np.random.default_rng(seed=2)
        int_categories = np_random.integers(low=1, high=20, size=nb_int_categories)
        str_categories = [
            ''.join(random.choices(string.ascii_uppercase + string.digits, k=random.randint(7, 8)))
            for category_i in range(nb_categories - nb_int_categories)
        ]
        categories = list(int_categories) + str_categories
        category_sizes = np_random.integers(low=1, high=max_category_size, size=nb_categories)
        category_indexes = np.array(range(nb_categories), dtype=int)
        sample_category_indexes = np.repeat(category_indexes, category_sizes)
        np_random.shuffle(sample_category_indexes)
        samples = [categories[i] for i in sample_category_indexes]
        data_size = len(samples)
        priming_data = samples
        test_data = []
        if(test_data_rel_size > 0.):
            test_data_size = round(data_size * test_data_rel_size) + 1
            test_data = priming_data[:test_data_size]
        return priming_data, test_data

    def create_test_data_old(self):
        random.seed(2)
        cateogries = [''.join(random.choices(string.ascii_uppercase + string.digits,
                              k=random.randint(7, 8))) for x in range(500)]
        for i in range(len(cateogries)):
            if i % 10 == 0:
                cateogries[i] = random.randint(1, 20)

        priming_data = []
        test_data = []
        for category in cateogries:
            times = random.randint(1, 50)
            for i in range(times):
                priming_data.append(category)
                if i % 3 == 0 or i == 1:
                    test_data.append(category)

        random.shuffle(priming_data)
        random.shuffle(test_data)
        return priming_data, test_data

    def test_autoencoder(self):
        """
        Checks reconstruction accuracy above 70% for a set of categories, length 8, for up to 500 unique categories (actual around 468).
        """  # noqa
        log.setLevel(logging.DEBUG)

        torch.manual_seed(2)

        priming_data, test_data = self.create_test_data()

        enc = CategoricalAutoEncoder(stop_after=20)
        enc.desired_error = 3

        enc.prepare(pd.Series(priming_data), pd.Series(priming_data))
        encoded_data = enc.encode(test_data)
        decoded_data = enc.decode(encoded_data)

        encoder_accuracy = accuracy_score(list(map(str, test_data)), list(map(str, decoded_data)))
        print(f'Categorical encoder accuracy for: {encoder_accuracy} on testing dataset')
        self.assertTrue(encoder_accuracy > 0.70)

    def check_encoder_on_device(self, device):
        priming_data, _ = self.create_test_data(nb_categories=8,
                                                nb_int_categories=3,
                                                max_category_size=3,
                                                test_data_rel_size=0.)

        enc = CategoricalAutoEncoder(stop_after=5, device=device)
        enc.prepare(pd.Series(priming_data), pd.Series(priming_data))
        self.assertEqual(list(enc.net.parameters())[0].device.type, device)

    def test_encoder_on_cpu(self):
        self.check_encoder_on_device('cpu')

    @unittest.skipIf(not torch.cuda.is_available(), 'CUDA unavailable')
    def test_encoder_on_cuda(self):
        self.check_encoder_on_device('cuda')



import unittest
import numpy as np
import pandas as pd
from sklearn.metrics import balanced_accuracy_score
from lightwood.api.types import ProblemDefinition
from lightwood.api.high_level import json_ai_from_problem, predictor_from_json_ai, JsonAI, code_from_json_ai, predictor_from_code  # noqa


np.random.seed(42)


class TestBasic(unittest.TestCase):

    def get_submodels(self):
        submodels = [
            {
                'module': 'RandomForest',
                'args': {
                    'stop_after': '$problem_definition.seconds_per_mixer',
                    'fit_on_dev': True,
                    'target': '$target',
                    'dtype_dict': '$dtype_dict',
                    'target_encoder': '$encoders[self.target]',
                    'use_optuna': True
                }
            },
        ]
        return submodels

    def test_0_regression(self):
        df = pd.read_csv('tests/data/concrete_strength.csv')[:500]
        target = 'concrete_strength'

        pdef = ProblemDefinition.from_dict({'target': target, 'time_aim': 80})
        jai = json_ai_from_problem(df, pdef)

        jai.model['args']['submodels'] = self.get_submodels()
        code = code_from_json_ai(jai)
        predictor = predictor_from_code(code)

        predictor.learn(df)
        predictor.predict(df)

    def test_1_binary(self):
        df = pd.read_csv('tests/data/ionosphere.csv')[:100]
        target = 'target'

        pdef = ProblemDefinition.from_dict({'target': target, 'time_aim': 20, 'unbias_target': False})
        jai = json_ai_from_problem(df, pdef)

        jai.model['args']['submodels'] = [
            {
                'module': 'RandomForest',
                'args': {'stop_after': '$problem_definition.seconds_per_mixer', 'fit_on_dev': True}}
        ]
        code = code_from_json_ai(jai)
        predictor = predictor_from_code(code)

        predictor.learn(df)
        predictions = predictor.predict(df)

        acc = balanced_accuracy_score(df[target], predictions['prediction'])
        self.assertTrue(acc > 0.5)
        self.assertTrue(all([0 <= p <= 1 for p in predictions['confidence']]))
